/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::ExampleMemory
 * @author     Rainer Kartmann ( rainer dot kartmann at kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ExampleMemory.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <SimoxUtility/algorithm/string.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/server/MemoryRemoteGui.h>

#include <RobotAPI/components/armem/server/ExampleMemory/aron/ExampleData.aron.generated.h>


namespace armarx
{
    armarx::PropertyDefinitionsPtr ExampleMemory::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs = new ComponentPropertyDefinitions(getConfigIdentifier());

        defs->topic(debugObserver);

        defs->optional(p.memoryName, "memory.Name", "Name of this memory (server).");

        p.core._defaultSegmentsStr = simox::alg::join(p.core.defaultCoreSegments, ", ");
        defs->optional(p.core._defaultSegmentsStr, "core.DefaultSegments",
                       "Core segments to add on start up (just as example).");
        defs->optional(p.core.addOnUsage, "core.AddOnUsage",
                       "If enabled, core segments are added when required by a new provider segment."
                       "This will usually be off for most memory servers.");

        return defs;
    }


    std::string ExampleMemory::getDefaultName() const
    {
        return "ExampleMemory";
    }

    void ExampleMemory::onInitComponent()
    {
        workingMemory.name() = p.memoryName;
        longtermMemory.name() = p.memoryName;

        // Usually, the memory server will specify a number of core segments with a specific aron type.
        workingMemory.addCoreSegment("ExampleData", armem::example::ExampleData::toAronType());

        // For illustration purposes, we add more segments (without types).
        bool trim = true;
        p.core.defaultCoreSegments = simox::alg::split(p.core._defaultSegmentsStr, ",", trim);
        p.core._defaultSegmentsStr.clear();
        workingMemory.addCoreSegments(p.core.defaultCoreSegments);
    }

    void ExampleMemory::onConnectComponent()
    {
        RemoteGui__createTab();
        RemoteGui_startRunningTask();
    }

    void ExampleMemory::onDisconnectComponent()
    {
    }

    void ExampleMemory::onExitComponent()
    {
    }



    // WRITING

    armem::data::AddSegmentsResult ExampleMemory::addSegments(const armem::data::AddSegmentsInput& input, const Ice::Current&)
    {
        // This function is overloaded to trigger the remote gui rebuild.
        armem::data::AddSegmentsResult result = ComponentPluginUser::addSegments(input, p.core.addOnUsage);
        tab.rebuild = true;
        return result;
    }


    armem::data::CommitResult ExampleMemory::commit(const armem::data::Commit& commit, const Ice::Current&)
    {
        // This function is overloaded to trigger the remote gui rebuild.
        armem::data::CommitResult result = ComponentPluginUser::commit(commit);
        tab.rebuild = true;
        return result;
    }


    // READING

    // Inherited from Plugin



    // REMOTE GUI

    void ExampleMemory::RemoteGui__createTab()
    {
        using namespace armarx::RemoteGui::Client;

        {
            std::scoped_lock lock(workingMemoryMutex);
            tab.memoryGroup = armem::server::MemoryRemoteGui().makeGroupBox(workingMemory);
        }

        VBoxLayout root = {tab.memoryGroup, VSpacer()};
        RemoteGui_createTab(getName(), root, &tab);
    }


    void ExampleMemory::RemoteGui_update()
    {
        if (tab.rebuild.exchange(false))
        {
            RemoteGui__createTab();
        }
    }

}
