#pragma once


// STD/STL
#include <optional>
#include <vector>

// RobotAPI
#include <RobotAPI/interface/armem/server/ReadingMemoryInterface.h>
#include <RobotAPI/interface/armem/server/StoringMemoryInterface.h>
#include <RobotAPI/libraries/armem/core/workingmemory/EntitySnapshot.h>
// #include <RobotAPI/libraries/armem/core/workingmemory/ice_conversions.h>

#include "Query.h"


namespace armarx::armem::client
{

    /**
     * @brief Reads data from a memory server.
     */
    class Reader
    {

    public:

        /**
         * @brief Construct a memory reader.
         * @param memory The memory proxy.
         */
        Reader(server::ReadingMemoryInterfacePrx memory = nullptr);

        void setReadingMemory(server::ReadingMemoryInterfacePrx memory);

        /// Perform a query.
        QueryResult query(const QueryInput& input) const;
        armem::query::data::Result query(const armem::query::data::Input& input) const;

        QueryResult query(armem::query::data::MemoryQueryPtr query, DataMode dataMode = DataMode::WithData) const;
        QueryResult query(const armem::query::data::MemoryQuerySeq& queries, DataMode dataMode = DataMode::WithData) const;

        QueryResult query(const QueryBuilder& queryBuilder) const;


        /**
         * @brief Qeury a specific set of memory IDs.
         *
         * Each ID can refer to an entity, a snapshot or an instance. When not
         * referring to an entity instance, the latest snapshot and first
         * instance will be queried, respectively.
         *
         * All memory IDs must refer to the memory this reader is reading from.
         * If an ID refers to another memory, the query will not find it and it
         * will not be part of the result.
         *
         * @param ids The entity, snapshot or instance IDs.
         * @param dataMode Whether to include instance data or just meta data.
         * @return The query result.
         */
        QueryResult
        queryMemoryIDs(const std::vector<MemoryID>& ids, DataMode dataMode = DataMode::WithData) const;


        /**
         * @brief Query the given snapshot and return the latest existing snapshot.
         * @param snapshotIDs The snapshot (or entity) IDs.
         * @return The latest snapshot contained in the query result, if any.
         */
        std::optional<wm::EntitySnapshot>
        getLatestSnapshotOf(const std::vector<MemoryID>& snapshotIDs) const;


        /**
         * @brief Get the latest snapshots under the given memory ID.
         * @param id A memory, core segment, provider segment or entity ID.
         * @param dataMode With or without data.
         * @return The query result.
         */
        QueryResult
        getLatestSnapshotsIn(const MemoryID& id, DataMode dataMode = DataMode::WithData) const;

        /**
         * @brief Get the latest snapshot under the given memory ID.
         * @param id A memory, core segment, provider segment or entity ID.
         * @param dataMode With or without data.
         * @return The latest contained snapshot, if any.
         */
        std::optional<wm::EntitySnapshot>
        getLatestSnapshotIn(const MemoryID& id, DataMode dataMode = DataMode::WithData) const;


        /**
         * @brief Get all latest snapshots in the memory.
         * @param dataMode With or without data.
         * @return The query result.
         */
        QueryResult
        getAllLatestSnapshots(DataMode dataMode = DataMode::WithData) const;


        /**
         * @brief Get the whole memory content.
         * @param dataMode With or without data.
         * @return The query result.
         */
        QueryResult
        getAll(DataMode dataMode = DataMode::WithData) const;


        //data::StoreResult readAndStore(data::StoreInputSeq& input);
        data::StoreResult readAndStore(const data::StoreInput& input) const;


        inline operator bool() const
        {
            return bool(memoryPrx);
        }

    public:

        server::ReadingMemoryInterfacePrx memoryPrx;

    };

}
