#include "ComponentPlugin.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/error.h>

#include "MemoryToIceAdapter.h"

//#include <RobotAPI/libraries/armem/core/io/diskWriter/NlohmannJSON/NlohmannJSONDiskWriter.h>


namespace armarx::armem::server::plugins
{
    ComponentPlugin::~ComponentPlugin()
    {}


    void ComponentPlugin::postCreatePropertyDefinitions(PropertyDefinitionsPtr& properties)
    {
        MemoryNameSystemComponentPlugin::postCreatePropertyDefinitions(properties);

        ComponentPluginUser& parent = this->parent<ComponentPluginUser>();
        properties->topic(memoryListener, parent.memoryListenerDefaultName);

        properties->optional(parent.longtermMemoryEnabled, "ltm.00_enabled");
        properties->optional(parent.longtermMemory.dbsettings.host, "ltm.10_host");
        properties->optional(parent.longtermMemory.dbsettings.port, "ltm.11_port");
        properties->optional(parent.longtermMemory.dbsettings.user, "ltm.20_user");
        properties->optional(parent.longtermMemory.dbsettings.password, "ltm.21_password");
        properties->optional(parent.longtermMemory.dbsettings.database, "ltm.22_database");
    }


    void ComponentPlugin::postOnConnectComponent()
    {
        ComponentPluginUser& parent = this->parent<ComponentPluginUser>();

        if (isMemoryNameSystemEnabled() && parent.memoryNameSystem)
        {
            registerMemory(parent);
        }
        parent.iceMemory.setMemoryListener(memoryListener);

        // establishing connection to ltm and mongodb
        if (parent.longtermMemoryEnabled)
        {
            parent.longtermMemory.reload();
        }
    }


    void ComponentPlugin::preOnDisconnectComponent()
    {
        ComponentPluginUser& parent = this->parent<ComponentPluginUser>();
        if (isMemoryNameSystemEnabled() && parent.memoryNameSystem)
        {
            removeMemory(parent);
        }
    }


    data::RegisterMemoryResult ComponentPlugin::registerMemory(ComponentPluginUser& parent)
    {
        MemoryID id = MemoryID().withMemoryName(parent.workingMemory.name());
        MemoryInterfacePrx proxy = MemoryInterfacePrx::checkedCast(parent.getProxy());
        ARMARX_CHECK_NOT_NULL(proxy);

        data::RegisterMemoryResult result;
        try
        {
            parent.memoryNameSystem.registerServer(id, proxy);
            result.success = true;
            ARMARX_DEBUG << "Registered memory server for " << id << " in the Memory Name System (MNS).";
        }
        catch (const armem::error::ServerRegistrationOrRemovalFailed& e)
        {
            result.success = false;
            result.errorMessage = e.what();
            ARMARX_WARNING << e.what();
        }
        return result;
    }


    data::RemoveMemoryResult ComponentPlugin::removeMemory(ComponentPluginUser& parent)
    {
        MemoryID id = MemoryID().withMemoryName(parent.workingMemory.name());

        data::RemoveMemoryResult result;
        try
        {
            parent.memoryNameSystem.removeServer(id);
            result.success = true;
            ARMARX_DEBUG << "Removed memory server for " << id << " from the Memory Name System (MNS).";
        }
        catch (const armem::error::ServerRegistrationOrRemovalFailed& e)
        {
            result.success = false;
            result.errorMessage = e.what();
            ARMARX_WARNING << e.what();
        }
        catch (const Ice::NotRegisteredException&)
        {
            // It's ok, the MNS is gone.
            result.success = false;
            result.errorMessage = "Memory Name System is gone.";
        }
        return result;
    }

}


namespace armarx::armem::server
{
    ComponentPluginUser::ComponentPluginUser()
    {
        addPlugin(plugin);
    }


    // WRITING
    data::AddSegmentsResult ComponentPluginUser::addSegments(const data::AddSegmentsInput& input, const Ice::Current&)
    {
        bool addCoreSegmentOnUsage = false;
        return addSegments(input, addCoreSegmentOnUsage);
    }

    data::AddSegmentsResult ComponentPluginUser::addSegments(const data::AddSegmentsInput& input, bool addCoreSegments)
    {
        std::scoped_lock lock(workingMemoryMutex);
        data::AddSegmentsResult result = iceMemory.addSegments(input, addCoreSegments);
        return result;
    }


    data::CommitResult ComponentPluginUser::commit(const data::Commit& commitIce, const Ice::Current&)
    {
        std::scoped_lock lock(workingMemoryMutex);
        return iceMemory.commit(commitIce);
    }


    // READING
    armem::query::data::Result ComponentPluginUser::query(const armem::query::data::Input& input, const Ice::Current&)
    {
        std::scoped_lock lock(workingMemoryMutex);
        return iceMemory.query(input);
    }

    // LTM LOADING
    data::StoreResult ComponentPluginUser::store(const data::StoreInput& input, const Ice::Current&)
    {
        std::scoped_lock lock(workingMemoryMutex, longtermMemoryMutex);
        return iceMemory.store(input);
    }


    // LTM STORING
    armem::query::data::Result ComponentPluginUser::load(const armem::query::data::Input& input, const Ice::Current&)
    {
        std::scoped_lock lock(longtermMemoryMutex);
        return iceMemory.load(input);
    }

}
